@@ -12,8 +12,8 @@
 // This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
 // WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 //
-// $URL$
-// $Id$
+// $URL: https://github.com/CGAL/cgal/blob/releases/CGAL-4.14.3/Mesh_3/include/CGAL/Mesh_3/Detect_polylines_in_polyhedra.h $
+// $Id: Detect_polylines_in_polyhedra.h 1e9dc49 2017-12-05T18:31:05+01:00 Mael Rouxel-Labbé
 // SPDX-License-Identifier: GPL-3.0+
 //
 //
@@ -25,19 +25,23 @@
 
 #include <CGAL/license/Mesh_3.h>
 
-
-#include <CGAL/Mesh_3/Detect_polylines_in_polyhedra_fwd.h>
 #include <CGAL/Compare_handles_with_or_without_timestamps.h>
+#include <CGAL/Mesh_3/Detect_polylines_in_polyhedra_fwd.h>
 #include <CGAL/Default.h>
+#include <CGAL/Hash_handles_with_or_without_timestamps.h>
 
-#include <algorithm>
 #include <boost/foreach.hpp>
 #include <boost/mpl/if.hpp>
+#include <boost/unordered_map.hpp>
+#include <boost/unordered_set.hpp>
+
+#include <algorithm>
 
 namespace CGAL { namespace Mesh_3 {
 
 template <typename Polyhedron>
-struct Detect_polylines {
+struct Detect_polylines
+{
   typedef typename Polyhedron::Traits Geom_traits;
   typedef typename Geom_traits::Point_3 Point_3;
   typedef typename Polyhedron::Halfedge_const_handle Halfedge_const_handle;
@@ -47,12 +51,13 @@ struct Detect_polylines {
   typedef typename Polyhedron::size_type size_type;
   typedef CGAL::Compare_handles_with_or_without_timestamps Compare_handles;
 
-  typedef std::set<Vertex_handle, Compare_handles> Vertices_set;
-  typedef std::map<Vertex_handle, 
-                   size_type,                    
-                   Compare_handles> Vertices_counter;
+  typedef CGAL::Hash_handles_with_or_without_timestamps   Hash_fct;
+  typedef boost::unordered_set<Vertex_handle, Hash_fct>   Vertices_set;
+  typedef boost::unordered_map<Vertex_handle,
+                               size_type,
+                               Hash_fct>                  Vertices_counter;
 
-  typedef std::set<Halfedge_handle, Compare_handles> Feature_edges_set;
+  typedef boost::unordered_set<Halfedge_handle, Hash_fct> Feature_edges_set;
 
   Feature_edges_set edges_to_consider;
   Vertices_set corner_vertices;
