@@ -13,8 +13,9 @@
 // This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
 // WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 //
-// $URL$
-// $Id$
+// $URL: https://github.com/CGAL/cgal/blob/releases/CGAL-4.14.3/Mesh_3/include/CGAL/Polyhedral_mesh_domain_with_features_3.h $
+// $Id: Polyhedral_mesh_domain_with_features_3.h dfb37e2 2019-06-07T12:06:44+02:00 Laurent Rineau
+// SPDX-License-Identifier: GPL-3.0+
 //
 //
 // Author(s)     : Laurent Rineau, Stéphane Tayeb
@@ -29,156 +30,42 @@
 
 #include <CGAL/license/Mesh_3.h>
 
+#include <CGAL/disable_warnings.h>
 
 #include <CGAL/Mesh_3/config.h>
 
-#include <CGAL/Random.h>
 #include <CGAL/Polyhedral_mesh_domain_3.h>
 #include <CGAL/Mesh_domain_with_polyline_features_3.h>
 #include <CGAL/Mesh_polyhedron_3.h>
-
 #include <CGAL/Mesh_3/Detect_polylines_in_polyhedra.h>
 #include <CGAL/Mesh_3/Polyline_with_context.h>
-#include <CGAL/Mesh_3/Detect_features_in_polyhedra.h>
 
-#include <CGAL/enum.h>
 #include <CGAL/IO/Polyhedron_iostream.h>
+#include <CGAL/internal/Mesh_3/helpers.h>
+
+#include <CGAL/boost/graph/split_graph_into_polylines.h>
+#include <CGAL/boost/iterator/transform_iterator.hpp>
 #include <CGAL/boost/graph/graph_traits_Polyhedron_3.h>
 #include <CGAL/boost/graph/helpers.h>
+#include <CGAL/enum.h>
+#include <CGAL/Default.h>
+#include <CGAL/Polygon_mesh_processing/detect_features.h>
+#include <CGAL/Random.h>
+
+#include <boost/foreach.hpp>
 #include <boost/graph/filtered_graph.hpp>
 #include <boost/graph/adjacency_list.hpp>
-#include <CGAL/boost/graph/split_graph_into_polylines.h>
 
-#include <boost/iterator/transform_iterator.hpp>
-#include <boost/foreach.hpp>
-
-#include <string>
-#include <vector>
 #include <fstream>
-
+#include <iostream>
+#include <map>
+#include <string>
+#include <set>
+#include <vector>
+#include <utility>
 
 namespace CGAL {
 
-namespace internal {
-namespace Mesh_3 {
-
-template <typename Kernel>
-struct Angle_tester
-{
-  template <typename vertex_descriptor, typename Graph>
-  bool operator()(vertex_descriptor& v, const Graph& g) const
-  {
-    typedef typename boost::graph_traits<Graph>::out_edge_iterator out_edge_iterator;
-    if (out_degree(v, g) != 2)
-      return true;
-    else
-    {
-      out_edge_iterator out_edge_it, out_edges_end;
-      boost::tie(out_edge_it, out_edges_end) = out_edges(v, g);
-
-      vertex_descriptor v1 = target(*out_edge_it++, g);
-      vertex_descriptor v2 = target(*out_edge_it++, g);
-      CGAL_assertion(out_edge_it == out_edges_end);
-
-      const typename Kernel::Point_3& p = g[v];
-      const typename Kernel::Point_3& p1 = g[v1];
-      const typename Kernel::Point_3& p2 = g[v2];
-
-      return (CGAL::angle(p1, p, p2) == CGAL::ACUTE);
-    }
-  }
-};
-
-template <typename Polyhedron>
-struct Is_featured_edge {
-  const Polyhedron* polyhedron;
-  Is_featured_edge() : polyhedron(0) {} // required by boost::filtered_graph
-  Is_featured_edge(const Polyhedron& polyhedron) : polyhedron(&polyhedron) {}
-
-  bool operator()(typename boost::graph_traits<Polyhedron>::edge_descriptor e) const {
-    return halfedge(e, *polyhedron)->is_feature_edge();
-  }
-}; // end Is_featured_edge<Polyhedron>
-
-template <typename Polyhedron>
-struct Is_border_edge {
-  const Polyhedron* polyhedron;
-  Is_border_edge() : polyhedron(0) {} // required by boost::filtered_graph
-  Is_border_edge(const Polyhedron& polyhedron) : polyhedron(&polyhedron) {}
-
-  bool operator()(typename boost::graph_traits<Polyhedron>::edge_descriptor e) const {
-    return is_border(halfedge(e, *polyhedron), *polyhedron) ||
-      is_border(opposite(halfedge(e, *polyhedron), *polyhedron), *polyhedron);
-  }
-}; // end Is_featured_edge<Polyhedron>
-
-template<typename Polyhedral_mesh_domain,
-         typename Polyline_with_context,
-         typename Graph>
-struct Extract_polyline_with_context_visitor
-{
-  typedef typename Polyhedral_mesh_domain::Polyhedron Polyhedron;
-  std::vector<Polyline_with_context>& polylines;
-  const Graph& graph;
-
-  Extract_polyline_with_context_visitor
-  (const Graph& graph,
-   typename std::vector<Polyline_with_context>& polylines)
-    : polylines(polylines), graph(graph)
-  {}
-
-  void start_new_polyline()
-  {
-    polylines.push_back(Polyline_with_context());
-  }
-
-  void add_node(typename boost::graph_traits<Graph>::vertex_descriptor vd)
-  {
-    if(polylines.back().polyline_content.empty()) {
-      polylines.back().polyline_content.push_back(graph[vd]);
-    }
-  }
-
-  void add_edge(typename boost::graph_traits<Graph>::edge_descriptor ed)
-  {
-    typename boost::graph_traits<Graph>::vertex_descriptor
-      s = source(ed, graph),
-      t = target(ed, graph);
-    Polyline_with_context& polyline = polylines.back();
-    CGAL_assertion(!polyline.polyline_content.empty());
-    if(polyline.polyline_content.back() != graph[s]) {
-      polyline.polyline_content.push_back(graph[s]);
-    } else if(polyline.polyline_content.back() != graph[t]) {
-      // if the edge is zero-length, it is ignored
-      polyline.polyline_content.push_back(graph[t]);
-    }
-    const typename boost::edge_bundle_type<Graph>::type &
-      set_of_indices = graph[ed];
-    polyline.context.adjacent_patches_ids.insert(set_of_indices.begin(),
-                                                 set_of_indices.end());
-  }
-
-  void end_polyline()
-  {
-    // ignore degenerated polylines
-    if(polylines.back().polyline_content.size() < 2)
-      polylines.resize(polylines.size() - 1);
-    // else {
-    //   std::cerr << "Polyline with " << polylines.back().polyline_content.size()
-    //             << " vertices, incident to "
-    //             << polylines.back().context.adjacent_patches_ids.size()
-    //             << " patches:\n ";
-    //   for(auto p: polylines.back().polyline_content)
-    //     std::cerr << " " << p;
-    //   std::cerr << "\n";
-    // }
-  }
-};
-
-
-} // end CGAL::internal::Mesh_3
-} // end CGAL::internal
-
 /**
  * @class Polyhedral_mesh_domain_with_features_3
  *
@@ -186,39 +73,49 @@ struct Extract_polyline_with_context_visitor
  */
 template < class IGT_,
            class Polyhedron_ = typename Mesh_polyhedron_3<IGT_>::type,
-           class TriangleAccessor=Triangle_accessor_3<Polyhedron_,IGT_>,
-           class Use_patch_id_tag = Tag_true,
+           class TriangleAccessor= CGAL::Default,
+           class Patch_id=int,
            class Use_exact_intersection_construction_tag = Tag_true >
 class Polyhedral_mesh_domain_with_features_3
   : public Mesh_domain_with_polyline_features_3<
       Polyhedral_mesh_domain_3< Polyhedron_,
                                 IGT_,
                                 TriangleAccessor,
-                                Use_patch_id_tag,
+                                Patch_id,
                                 Use_exact_intersection_construction_tag > >
 {
   typedef Mesh_domain_with_polyline_features_3<
     Polyhedral_mesh_domain_3<
       Polyhedron_, IGT_, TriangleAccessor,
-      Use_patch_id_tag, Use_exact_intersection_construction_tag > > Base;
+      Patch_id, Use_exact_intersection_construction_tag > > Base;
 
   typedef boost::adjacency_list<
     boost::setS, // this avoids parallel edges
     boost::vecS,
     boost::undirectedS,
-    typename Polyhedron_::Point,
-    typename Polyhedron_::Vertex::Set_of_indices> Featured_edges_copy_graph;
-
+    typename IGT_::Point_3,
+    std::set<typename Base::Surface_patch_index> > Featured_edges_copy_graph;
 public:
   typedef Polyhedron_ Polyhedron;
+  typedef Polyhedron Polyhedron_type;
 
   // Index types
   typedef typename Base::Index                Index;
   typedef typename Base::Corner_index         Corner_index;
-  typedef typename Base::Curve_segment_index  Curve_segment_index;
+  typedef typename Base::Curve_index          Curve_index;
   typedef typename Base::Surface_patch_index  Surface_patch_index;
   typedef typename Base::Subdomain_index      Subdomain_index;
-  
+
+#ifndef CGAL_NO_DEPRECATED_CODE
+  typedef Curve_index Curve_segment_index; ///< Backward-compatibility
+#endif
+
+  typedef typename boost::property_map<Polyhedron,
+                                       face_patch_id_t<Patch_id>
+                                       >::type            Face_patch_id_pmap;
+  typedef typename boost::property_traits<
+    Face_patch_id_pmap>::value_type                       P_id;
+
   // Backward compatibility
 #ifndef CGAL_MESH_3_NO_DEPRECATED_SURFACE_INDEX
   typedef Surface_patch_index                 Surface_index;
@@ -227,11 +124,11 @@ public:
   typedef typename Base::R         R;
   typedef typename Base::Point_3   Point_3;
   typedef typename Base::FT        FT;
-  
+
   typedef CGAL::Tag_true           Has_features;
 
   typedef std::vector<Point_3> Bare_polyline;
-  typedef Mesh_3::Polyline_with_context<Surface_patch_index, Curve_segment_index,
+  typedef Mesh_3::Polyline_with_context<Surface_patch_index, Curve_index,
                                         Bare_polyline > Polyline_with_context;
   /// Constructors
   Polyhedral_mesh_domain_with_features_3(const Polyhedron& p,
@@ -240,10 +137,13 @@ public:
   {
     stored_polyhedra.resize(1);
     stored_polyhedra[0] = p;
+    get(face_patch_id_t<Patch_id>(), stored_polyhedra[0]);
     this->add_primitives(stored_polyhedra[0]);
     this->build();
   }
 
+#ifndef CGAL_NO_DEPRECATED_CODE
+
   CGAL_DEPRECATED
   Polyhedral_mesh_domain_with_features_3(const std::string& filename,
                                          CGAL::Random* p_rng = NULL)
@@ -262,6 +162,7 @@ public:
   {
     load_from_file(filename);
   }
+#endif // not CGAL_NO_DEPRECATED_CODE
 
   Polyhedral_mesh_domain_with_features_3(const Polyhedron& p,
                                          const Polyhedron& bounding_p,
@@ -271,9 +172,15 @@ public:
     stored_polyhedra.resize(2);
     stored_polyhedra[0] = p;
     stored_polyhedra[1] = bounding_p;
+    get(face_patch_id_t<Patch_id>(), stored_polyhedra[0]);
+    get(face_patch_id_t<Patch_id>(), stored_polyhedra[1]);
     this->add_primitives(stored_polyhedra[0]);
     this->add_primitives(stored_polyhedra[1]);
-    this->add_primitives_to_bounding_tree(stored_polyhedra[1]);
+    if(CGAL::is_empty(bounding_p)) {
+      this->set_surface_only();
+    } else {
+      this->add_primitives_to_bounding_tree(stored_polyhedra[1]);
+    }
   }
 
   template <typename InputPolyhedraPtrIterator>
@@ -285,6 +192,7 @@ public:
     stored_polyhedra.reserve(std::distance(begin, end));
     for (; begin != end; ++begin) {
       stored_polyhedra.push_back(**begin);
+      get(face_patch_id_t<Patch_id>(), stored_polyhedra.back());
       this->add_primitives(stored_polyhedra.back());
     }
     this->set_surface_only();
@@ -302,12 +210,18 @@ public:
     if(begin != end) {
       for (; begin != end; ++begin) {
         stored_polyhedra.push_back(**begin);
+        get(face_patch_id_t<Patch_id>(), stored_polyhedra.back());
         this->add_primitives(stored_polyhedra.back());
       }
-      stored_polyhedra.push_back(bounding_polyhedron);
-      this->add_primitives(stored_polyhedra.back());
     }
-    this->add_primitives_to_bounding_tree(stored_polyhedra.back());
+    stored_polyhedra.push_back(bounding_polyhedron);
+    get(face_patch_id_t<Patch_id>(), stored_polyhedra.back());
+    this->add_primitives(stored_polyhedra.back());
+    if(CGAL::is_empty(bounding_polyhedron)) {
+      this->set_surface_only();
+    } else {
+      this->add_primitives_to_bounding_tree(stored_polyhedra.back());
+    }
     this->build();
   }
 
@@ -318,17 +232,27 @@ public:
   void initialize_ts(Polyhedron& p);
 
   void detect_features(FT angle_in_degree, std::vector<Polyhedron>& p);
-  void detect_features(FT angle_in_degree = FT(60)) { detect_features(angle_in_degree, stored_polyhedra); }
+  void detect_features(FT angle_in_degree = FT(60))
+  {
+    detect_features(angle_in_degree, stored_polyhedra);
+  }
 
   void detect_borders(std::vector<Polyhedron>& p);
   void detect_borders() { detect_borders(stored_polyhedra); };
 
+  // non-documented, provided to the FEniCS project
+  const std::vector<Polyhedron>& polyhedra()const
+  {
+    return stored_polyhedra;
+  }
+
 private:
   void load_from_file(const char* filename) {
     // Create input polyhedron
     std::ifstream input(filename);
     stored_polyhedra.resize(1);
     input >> stored_polyhedra[0];
+    get(face_patch_id_t<Patch_id>(), stored_polyhedra[0]);
     this->add_primitives(stored_polyhedra[0]);
     this->build();
   }
@@ -359,21 +283,27 @@ void
 Polyhedral_mesh_domain_with_features_3<GT_,P_,TA_,Tag_,E_tag_>::
 initialize_ts(Polyhedron& p)
 {
+  typedef typename boost::property_map<Polyhedron,vertex_time_stamp_t>::type Vtmap;
+  typedef typename boost::property_map<Polyhedron,halfedge_time_stamp_t>::type Htmap;
+  typedef typename boost::property_map<Polyhedron,face_time_stamp_t>::type Ftmap;
+  Vtmap vtm = get(vertex_time_stamp,p);
+  Htmap htm = get(halfedge_time_stamp,p);
+  Ftmap ftm = get(face_time_stamp,p);
+
   std::size_t ts = 0;
-  for(typename Polyhedron::Vertex_iterator v = p.vertices_begin(),
-      end = p.vertices_end() ; v != end ; ++v)
+  BOOST_FOREACH(typename boost::graph_traits<Polyhedron>::vertex_descriptor vd, vertices(p))
   {
-    v->set_time_stamp(ts++);
+    put(vtm,vd,ts++);
   }
-  for(typename Polyhedron::Facet_iterator fit = p.facets_begin(),
-       end = p.facets_end() ; fit != end ; ++fit )
+
+  BOOST_FOREACH(typename boost::graph_traits<Polyhedron>::face_descriptor fd, faces(p))
   {
-    fit->set_time_stamp(ts++);
+    put(ftm,fd,ts++);
   }
-  for(typename Polyhedron::Halfedge_iterator hit = p.halfedges_begin(),
-       end = p.halfedges_end() ; hit != end ; ++hit )
+
+  BOOST_FOREACH(typename boost::graph_traits<Polyhedron>::halfedge_descriptor hd, halfedges(p))
   {
-    hit->set_time_stamp(ts++);
+    put(htm,hd,ts++);
   }
 }
 
@@ -415,27 +345,41 @@ detect_features(FT angle_in_degree, std::vector<Polyhedron>& poly)
   typedef Featured_edges_copy_graph G_copy;
   G_copy g_copy;
   typedef typename boost::graph_traits<G_copy>::vertex_descriptor vertex_descriptor;
-  typedef std::map<typename Polyhedron::Point,
+  typedef typename boost::property_traits<typename boost::property_map<Polyhedron,vertex_point_t>::type>::value_type Point;
+  typedef std::map<Point,
                    vertex_descriptor> P2vmap;
   // TODO: replace this map by and unordered_map
   P2vmap p2vmap;
-
-  CGAL::Mesh_3::Detect_features_in_polyhedra<Polyhedron> detect_features;
+  namespace PMP = CGAL::Polygon_mesh_processing;
+  std::size_t nb_of_patch_plus_one = 1;
   BOOST_FOREACH(Polyhedron& p, poly)
   {
     initialize_ts(p);
+    typedef typename boost::property_map<Polyhedron,CGAL::face_patch_id_t<Tag_> >::type PIDMap;
+    typedef typename boost::property_map<Polyhedron,CGAL::vertex_incident_patches_t<P_id> >::type VIPMap;
+    typedef typename boost::property_map<Polyhedron, CGAL::edge_is_feature_t>::type EIFMap;
+
+    using Mesh_3::internal::Get_face_index_pmap;
+    Get_face_index_pmap<Polyhedron> get_face_index_pmap(p);
+
+    PIDMap pid_map = get(face_patch_id_t<Tag_>(), p);
+    VIPMap vip_map = get(vertex_incident_patches_t<P_id>(), p);
+    EIFMap eif_map = get(CGAL::edge_is_feature, p);
 
     // Get sharp features
-    detect_features.detect_sharp_edges(p, angle_in_degree);
-    detect_features.detect_surface_patches(p);
-    detect_features.detect_vertices_incident_patches(p);
+    nb_of_patch_plus_one += PMP::sharp_edges_segmentation(p, angle_in_degree
+      , eif_map
+      , pid_map
+      , PMP::parameters::first_index(nb_of_patch_plus_one)
+      .face_index_map(get_face_index_pmap(p))
+      .vertex_incident_patches_map(vip_map));
 
-    internal::Mesh_3::Is_featured_edge<Polyhedron> is_featured_edge(p);
+    Mesh_3::internal::Is_featured_edge<Polyhedron> is_featured_edge(p);
 
     add_featured_edges_to_graph(p, is_featured_edge, g_copy, p2vmap);
   }
   add_features_from_split_graph_into_polylines(g_copy);
-  borders_detected_ = true;/*done by Mesh_3::detect_features*/
+  borders_detected_ = true;/*done by PMP::detect_features*/
 }
 
 template < typename GT_, typename P_, typename TA_,
@@ -460,12 +404,11 @@ add_features_from_split_graph_into_polylines(Featured_edges_copy_graph& g_copy)
 {
   std::vector<Polyline_with_context> polylines;
 
-  internal::Mesh_3::Extract_polyline_with_context_visitor<
-    Polyhedral_mesh_domain_with_features_3,
+  Mesh_3::internal::Extract_polyline_with_context_visitor<
     Polyline_with_context,
     Featured_edges_copy_graph
     > visitor(g_copy, polylines);
-  internal::Mesh_3::Angle_tester<GT_> angle_tester;
+  Mesh_3::internal::Angle_tester<GT_> angle_tester;
   split_graph_into_polylines(g_copy, visitor, angle_tester);
 
   this->add_features_with_context(polylines.begin(),
@@ -511,29 +454,34 @@ add_featured_edges_to_graph(const Polyhedron& p,
 
   const Featured_edges_graph& graph = orig_graph;
 
+  typedef typename boost::property_map<Polyhedron,vertex_point_t>::const_type Vpm;
+  Vpm vpm = get(vertex_point, p);
   BOOST_FOREACH(Graph_vertex_descriptor v, vertices(graph)){
     vertex_descriptor vc;
-    typename P2vmap::iterator it = p2vmap.find(v->point());
+    typename P2vmap::iterator it = p2vmap.find(get(vpm,v));
     if(it == p2vmap.end()) {
       vc = add_vertex(g_copy);
-      g_copy[vc] = v->point();
-      p2vmap[v->point()] = vc;
+      g_copy[vc] = get(vpm, v);
+      p2vmap[get(vpm,v)] = vc;
     }
   }
 
+  typedef typename boost::property_map<Polyhedron,face_patch_id_t<Tag_> >::type Face_patch_id_pmap;
+  Face_patch_id_pmap fpm = get(face_patch_id_t<Tag_>(),p);
+
   BOOST_FOREACH(Graph_edge_descriptor e, edges(graph)){
-    vertex_descriptor vs = p2vmap[source(e,graph)->point()];
-    vertex_descriptor vt = p2vmap[target(e,graph)->point()];
+    vertex_descriptor vs = p2vmap[get(vpm,source(e,graph))];
+    vertex_descriptor vt = p2vmap[get(vpm,target(e,graph))];
     CGAL_warning_msg(vs != vt, "ignore self loop");
     if(vs != vt) {
       const std::pair<edge_descriptor, bool> pair = add_edge(vs,vt,g_copy);
-      typename Polyhedron::Halfedge_handle he = halfedge(e, p);
+      typename boost::graph_traits<Polyhedron>::halfedge_descriptor he = halfedge(e, p);
       if(!is_border(he, p)) {
-        g_copy[pair.first].insert(he->face()->patch_id());;
+        g_copy[pair.first].insert(get(fpm, face(he,p)));
       }
-      he = he->opposite();
+      he = opposite(he,p);
       if(!is_border(he, p)) {
-        g_copy[pair.first].insert(he->face()->patch_id());;
+        g_copy[pair.first].insert(get(fpm, face(he,p)));
       }
     }
   }
@@ -548,5 +496,6 @@ add_featured_edges_to_graph(const Polyhedron& p,
 
 } //namespace CGAL
 
+#include <CGAL/enable_warnings.h>
 
 #endif // CGAL_POLYHEDRAL_MESH_DOMAIN_WITH_FEATURES_3_H
